import {
    CustomViewInterface,
    FeatureBlockInterface,
    FeatureViewInterface
} from "../RcsbFvSequence/SequenceViews/CustomView/CustomView";
import {
    RcsbFv,
    RcsbFvDisplayTypes,
    RcsbFvRowConfigInterface,
    RcsbFvTrackDataElementInterface
} from "@rcsb/rcsb-saguaro";
import { RegionSelectionInterface } from "../RcsbFvState/RcsbFvSelectorManager";
import {
    StructureViewerPublicInterface, SaguaroRegionList
} from "../RcsbFvStructure/StructureViewerInterface";
import {
    LoadMethod,
    LoadMolstarInterface,
    LoadMolstarReturnType
} from "../RcsbFvStructure/StructureViewers/MolstarViewer/MolstarActionManager";
import {RcsbFvStateManager} from "../RcsbFvState/RcsbFvStateManager";

import {RcsbFvStructureConfigInterface} from "../RcsbFvStructure/RcsbFvStructure";
import {ViewerProps} from "@rcsb/rcsb-molstar/build/src/viewer";
import { updateSiteColors } from "./tmdet-extension/tmdet-color-theme";
import { TmDetRcsbPreset } from "./TmTrajectoryHierarchyPreset";
import { TmRowTitleComponent } from "./tmdet-viewer/TmRowTitleComponent";

export type TmFvConfigParams = {
    pdbEntry: string,
    pdbtmEntry: string,
    htpEntry: string,
    labelAsymId: string,
    side1: string,
    title: string,
    subtitle?: string,
    fvConfigData?: any
};
export async function createFeatureViewerConfing(params: TmFvConfigParams) {

    const modelId = `${params.pdbEntry}_model`;

    const fvConfig: FeatureViewInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType> =
        params.fvConfigData.blockConfig[0].featureViewConfig[0];

    setFvConfigCallbacks(fvConfig, modelId, params.labelAsymId);

    const customConfig: CustomViewInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType> =
        params.fvConfigData;

    const sequenceConfig = {
        title: params.title,
        subtitle: params.subtitle ?? '',
        config: customConfig
    };

    const molstarConfig: RcsbFvStructureConfigInterface<LoadMolstarInterface<unknown,unknown>,{viewerProps:Partial<ViewerProps>}> = {
        loadConfig: {
            loadMethod: LoadMethod.loadStructureFromUrl,
            loadParams: {
                url: `https://www.ebi.ac.uk/pdbe/entry-files/download/${params.pdbEntry}_updated.cif`,
                format: 'mmcif',
                isBinary: false,
                id: `${params.pdbEntry}_model`,
                reprProvider: TmDetRcsbPreset
            }
        },
        structureViewerConfig: {
            viewerProps:{
                showImportControls: true,
                showSessionControls: false
            }
        }
    };

    updateSiteColors(params.side1 as any);

    setExternalRowTitleComponent(fvConfig.rowConfig, structureRowClickHandler);

    return { sequenceConfig: sequenceConfig, molstarConfig: molstarConfig };
}

function setFvConfigCallbacks(fvConfig: FeatureViewInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType>, modelId: string, labelAsymId: string) {
    fvConfig.sequenceSelectionChangeCallback = (plugin: StructureViewerPublicInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType>, stateManager: RcsbFvStateManager, sequenceRegion: Array<RcsbFvTrackDataElementInterface>) => {
        stateManager.selectionState.clearSelection("select", {modelId: modelId, labelAsymId: labelAsymId});
        if(sequenceRegion.length > 0) {
            const regions = sequenceRegion.map(r => ({
                modelId: modelId,
                labelAsymId: labelAsymId,
                region: {begin: r.begin, end: r.end ?? r.begin, source: "sequence"} as RegionSelectionInterface
            }));
            stateManager.selectionState.addSelectionFromMultipleRegions(regions, "select");
            plugin.select(regions.map(r => ({
                ...r,
                begin: r.region.begin,
                end: r.region.end
            })), "select", "set");
        }else{
            plugin.clearSelection("select", {modelId: modelId, labelAsymId: labelAsymId})
            plugin.resetCamera();
        }
    };
    fvConfig.sequenceElementClickCallback = (plugin: StructureViewerPublicInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType>, stateManager: RcsbFvStateManager, d: RcsbFvTrackDataElementInterface) => {
        if(d!=null)
            plugin.cameraFocus(modelId, labelAsymId, d.begin, d.end ?? d.begin);
    };
    fvConfig.sequenceHoverCallback = (plugin: StructureViewerPublicInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType>, stateManager: RcsbFvStateManager, elements: Array<RcsbFvTrackDataElementInterface>) => {
        if(elements == null || elements.length == 0)
            plugin.clearSelection("hover");
        else {
            const e = elements[0];
            plugin.select(elements.map(e=>({
                modelId: modelId,
                labelAsymId: labelAsymId,
                begin: e.begin,
                end: e.end ?? e.begin
            })), "hover", "set");
        }
    };
    fvConfig.structureSelectionCallback = (plugin: StructureViewerPublicInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType>, pfv: RcsbFv, stateManager: RcsbFvStateManager) => {
        const sel: SaguaroRegionList | undefined = stateManager.selectionState.getSelectionWithCondition(modelId, labelAsymId, "select");
        if(sel == null) {
            pfv.clearSelection("select");
            plugin.resetCamera();
        }else {
            pfv.setSelection({elements: sel.regions, mode: "select"});
        }
    };
    fvConfig.structureHoverCallback = (plugin: StructureViewerPublicInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType>, pfv: RcsbFv, stateManager: RcsbFvStateManager) => {
        const sel: SaguaroRegionList | undefined = stateManager.selectionState.getSelectionWithCondition(modelId, labelAsymId, "hover");
        if(sel == null)
            pfv.clearSelection("hover");
        else
            pfv.setSelection({elements:sel.regions, mode:"hover"});
    };

}

function setExternalRowTitleComponent(rows: RcsbFvRowConfigInterface[], titleClick: (e: React.MouseEvent) => void) {
    rows.forEach(track => {
        if (track.externalRowTitle && (track.externalRowTitle.rowTitleComponent as unknown) == "TmRowTitleComponent") {
            track.externalRowTitle.rowTitleComponent = TmRowTitleComponent;
            if (!track.externalRowTitle.rowTitleAdditionalProps) {
                track.externalRowTitle.rowTitleAdditionalProps = {};
            }
            // TODO: (track.externalRowTitle.rowTitleAdditionalProps as any).titleClick = titleClick;
        }
    });
}

function structureRowClickHandler(e: React.MouseEvent) {
    console.warn('TODO: implement structureRowClickHandler()');
}
